// Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
// SPDX-License-Identifier: Apache-2.0
mod dynamic_record_sizing;
mod io_behaviors;
mod prefer_low_latency;
mod record_padding;
mod serialization;
